/*
 "KiPad" macro keypad for KiCad
 Copyright Jan. 2021 Michael Ruppe, Core Electronics
 Licensed Creative Commons 4.0 Noncommercial Share-Alike
 Based off the work of Zack Freedman: https://github.com/ZackFreedman/Big-Wheel

 If all you want to do is re-map the function of the keypad / encoders then
 modify the contents of actions.h

 Intended to run on a Teensy3.2, likely works on Teensy LC and 4.0
 REMEMBER: Tools --> USB Type --> Serial + Keyboard + Mouse + Joystick
*/

#include "pins.h"
#include "actions.h"
#include <Encoder.h>

// #define DEBUG_DUMP_MATRIX
// #define DEBUG_DUMP_EVENTS

#define MAX_ACTIONS 6

Encoder leftKnob(leftKnobA, leftKnobB);
Encoder middleKnob(middleKnobA, middleKnobB);
Encoder rightKnob(rightKnobA, rightKnobB);

long lastKnobPositions[] = {0, 0, 0};

bool lastSwitchStates[numSwitches];
bool debouncedSwitchStates[numSwitches];
bool lastDebouncedSwitchStates[numSwitches];
elapsedMillis switchDebounceTimestamps[numSwitches];
const unsigned int switchDebounceTime = 50; // [ms]
bool lockKeysLatched = false;  // After locking/unlocking, both keys are disabled until both are released

int lastActions[MAX_ACTIONS];
int allActionsThisFrame[MAX_ACTIONS];
byte actionIndex = 0;


void setup() {
  Serial.begin(9600);
  // pinMode(LED_BUILTIN, OUTPUT);
  // digitalWrite(LED_BUILTIN, HIGH);

  // Prepare key matrix pins
  for (int i = 0; i < 4; i++) {
    pinMode(allButtonColumnPins[i], OUTPUT);
    digitalWrite(allButtonColumnPins[i], HIGH);
    pinMode(allButtonRowPins[i], INPUT_PULLUP);
  }

  // Encoder buttons - Encoder library prepares its own pins.
  pinMode(leftEncoderButton, INPUT_PULLUP);
  pinMode(middleEncoderButton, INPUT_PULLUP);
  pinMode(rightEncoderButton, INPUT_PULLUP);

}

void loop() {

  long knobPositions[] = {leftKnob.read(), middleKnob.read(), rightKnob.read()};

  bool switchStates[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  for (int i = 0; i < MAX_ACTIONS; i++) allActionsThisFrame[i] = NO_ACTION;
  actionIndex = 0;

  // Read keypad states into state-array
  for (int j = 0; j < 4; j++) { // Step through each column
      digitalWrite(allButtonColumnPins[j], LOW);
      for (int i = 0; i < 4; i++) { // step through each column
        if (!digitalRead(allButtonRowPins[i])) {
  #ifdef DEBUG_DUMP_MATRIX
          Serial.printf("%d,%d (Button %d)\n", i, j, keyMatrix[i][j]);

  #endif

          switchStates[keyMatrix[i][j]] = true;
        }
      }
      digitalWrite(allButtonColumnPins[j], HIGH);

  }

  // Read encoder button states into state-array
  for (int i = 0; i < numEncoders; i++){
    switchStates[encoderButtons[i]] = !digitalRead(allEncoderButtonPins[i]);
  }

  // Debounce all switches (Keypad, encoder button) - naughty polling based interrupts because I'm lazy
  for (int i = 0; i < numSwitches; i++) {
    // reset timer on edge
    if (switchStates[i] != lastSwitchStates[i]) {
      switchDebounceTimestamps[i] = 0;
    }
    // set state after timer burndown
    if (switchDebounceTimestamps[i] >= switchDebounceTime) {
      lastDebouncedSwitchStates[i] = debouncedSwitchStates[i];
      debouncedSwitchStates[i] = switchStates[i];
    }

  // Perform most actions only on falling edge
  if (debouncedSwitchStates[i] && !lastDebouncedSwitchStates[i]) {
    // Catch alternate functions. These are invoked by holding the KiPad's Ctrl key. Requires the program to have caught the held key and looped back around. Whatever.
#ifdef DEBUG_DUMP_EVENTS
      Serial.print("Queuing action ");
      Serial.println(switchAssignments[i]);
#endif
      queueAction(switchAssignments[i]);
    }
  }


  // clear holdable keys on release
  for (int i = 0; i < 4; i++) {
    for (int j = 0; j < 4; j++) {
      int code = keyMatrix[i][j];
      if (!switchStates[code] &&              // key released: react to key-release immediately (keypress was already debounced)
          lastDebouncedSwitchStates[code] &&  // key used to be held
          holdableKeys[i][j]){                // key is defined as 'holdable'
        Keyboard.release(switchAssignments[code]);

        // ugly edge-case. sorry. I want to be able to hold Ctrl+Z or Ctrl+Y for long undo/redo chains...
        if (switchAssignments[code] == ACT_UNDO_REDO){
          Keyboard.release(KEY_Z);
          Keyboard.release(KEY_Y);
          Keyboard.release(MODIFIERKEY_CTRL);
        }

        // also reset any state variables
        switch (switchAssignments[code]) {
          case MODIFIERKEY_CTRL:
            ctrlDown = false;
            break;
          case MODIFIERKEY_SHIFT:
            shiftDown = false;
            break;
          default:
            break;
        }
      }
    }
  }


  // Encoders
  for (int i = 0; i < 3; i++) {
      if (abs(knobPositions[i] - lastKnobPositions[i]) >= 4) {
#ifdef DEBUG_DUMP_EVENTS
        Serial.print("Knob ");
        Serial.print(i);
        Serial.print(" ");
        Serial.print(lastKnobPositions[i]);
        Serial.print(" --> ");
        Serial.print(knobPositions[i]);
        Serial.print(" ");
#endif

        if (knobPositions[i] > lastKnobPositions[i]) {
#ifdef DEBUG_DUMP_EVENTS
          Serial.println("CW");
#endif
          queueAction(knobAssignments[i][1]);
        }
        else {
#ifdef DEBUG_DUMP_EVENTS
          Serial.println("CCW");
#endif
          queueAction(knobAssignments[i][0]);
        }

        lastKnobPositions[i] = knobPositions[i];
      }
    }



  // Send the queued actions. Block double-sending the same action eg. held ctrl/shift keys
  for (int i = 0; i < MAX_ACTIONS; i++) {
    bool actionAlreadyDone = false;
    if (allActionsThisFrame[i] == NO_ACTION) continue;
    for (int j = 0; j < MAX_ACTIONS; j++) {
      if (lastActions[j] == allActionsThisFrame[i]) {
        actionAlreadyDone = true;
        break;
      }
    }
    if (!actionAlreadyDone) {
      performAction(allActionsThisFrame[i]);
    }
  }


  // Update states for the next loop
  for (int i = 0; i < numSwitches; i++)
    lastSwitchStates[i] = switchStates[i];

  for (int i = 0; i < MAX_ACTIONS; i++)
    lastActions[i] = allActionsThisFrame[i];

}



void queueAction(int action) {
#ifdef DEBUG_DUMP_EVENTS
  Serial.printf("Action: %d", action);
#endif

  if (action != NO_ACTION && actionIndex < MAX_ACTIONS -1) {
    allActionsThisFrame[actionIndex++] = action;
  }
}
